package ru.ifmo.qbeing.global;

import ru.ifmo.qbeing.*;
import ru.ifmo.qbeing.common.*;
import universum.bi.*;

import java.util.*;

public class GlobalStrategy {
    protected static GlobalStrategy instance;

    protected ColonyControl colonyControl;

    protected int beingCount;
    protected Map<BeingType, Integer> beingTypeCount;
    protected Map<BeingType, Set<QBeing>> beingsByType;

    protected boolean exploringFinished = false;
    private Object owner;
    private int maxTurns;
    private int maxExplorerSteps;
    private int lastKnownLocationsUpdatedTurn;
    private float exploredRate;
    public boolean heal;

    private StrategyAutomata automata;

    public GlobalStrategy() {
        automata = new StrategyAutomata(this);
        colonyControl = new ColonyControl();
        beingsByType = new HashMap<BeingType, Set<QBeing>>();
        beingCount = 0;
        beingTypeCount = new HashMap<BeingType, Integer>();
    }

    public static synchronized void initInstance() {
        instance = new GlobalStrategy();
    }

    public static synchronized GlobalStrategy instance() {
        if (instance == null) {
            initInstance();
        }
        return instance;
    }

    public void checkExploringFinished(BeingInterface bi) {
        if (exploringFinished)
            return;
        if (bi.getTurnsCount() > maxExplorerSteps || UniverseInfo.instance().getExploredLocationCount() >= getExploredRate() * Constants.getWidth() * Constants.getHeight()) {
            convertExplorersToEaters(bi);
            exploringFinished = true;
        }
    }

    private void convertExplorersToEaters(BeingInterface bi) {
        Set<QBeing> beingsSet = getBeingsByType(BeingType.TYPE_EXPLORER);
        QBeing[] beings = beingsSet.toArray(new QBeing[beingsSet.size()]);
        for (QBeing being : beings) {
            changeState(being.getState(), BeingState.S_EATER_FIND_FEED_PLACE, being, bi);
            EventManager.handleBeingEvent(BeingEvent.E107, being, new StateMachineContext(being, bi));
        }
    }

    private Integer getBeingTypeCount(BeingType type) {
        Integer i = beingTypeCount.get(type);
        if (i == null)
            i = 0;
        return i;
    }

    private Set<QBeing> getBeingsByType(BeingType type) {
        Set<QBeing> beings = beingsByType.get(type);
        if (beings == null) {
            beings = new HashSet<QBeing>();
            beingsByType.put(type, beings);
        }
        return beings;
    }

    public synchronized int getBeingsByTypeCount(BeingType type) {
        return getBeingsByType(type).size();
    }

    private void incBeingTypeCount(QBeing me, int state) {
        BeingType type = getType(state);
        beingTypeCount.put(type, getBeingTypeCount(type) + 1);
        getBeingsByType(type).add(me);
    }

    public static BeingType getType(int state) {
        switch(state) {
            case BeingState.S_EATER_FIND_FEED_PLACE:
            case BeingState.S_EATER_MOVING_TO_POINT:
            case BeingState.S_PRODUCER:
                return BeingType.TYPE_EATER;
            case BeingState.S_EXPLORING:
            case BeingState.S_EXPLORING_HUNGRY:
                return BeingType.TYPE_EXPLORER;
        }
        return null;
    }

    private void decBeingTypeCount(QBeing me, int state) {
        if (state != 0) {
            BeingType type = getType(state);
            beingTypeCount.put(type, getBeingTypeCount(type) - 1);
            getBeingsByType(type).remove(me);
        }
    }

    public synchronized void beingDead(QBeing me) {
        UniverseInfo.instance().removeBeingFromLocation(me, me.getLocation());
        decBeingTypeCount(me, me.getState());
        beingCount--;
    }

    public boolean startColonyHere(PointInfo pi, QBeing me) {
        return colonyControl.setupColony(pi, me);
    }

    public synchronized void colonySuplierRemoved(Location loc, QBeing me) {
        colonyControl.removeColonySupplier(loc, me);
    }

    private void setColonySupplier(QBeing me, BeingInterface bi) {
        colonyControl.addColonySupplier(me, bi);
    }

    public synchronized void setOwner(Object owner) {
        if (this.owner == null)
            this.owner = owner;
    }

    public synchronized Object getOwner() {
        return owner;
    }

    public synchronized void checkStep(BeingInterface bi) {
        checkGameState(bi, bi.getTurnsCount());
    }

    private void checkGameState(BeingInterface bi, int turnsCount) {
        if (turnsCount >= maxExplorerSteps) {
            convertExplorersToEaters(bi);
            exploringFinished = true;
        }
        if (lastKnownLocationsUpdatedTurn != turnsCount) {
            UniverseInfo.instance().updateEnergyAtKnownLocations();
            lastKnownLocationsUpdatedTurn = turnsCount;
        }
    }

    public synchronized void initGameInfo(UserGameInfo info) {
        maxTurns = info.maxTurns;
        switch (info.kind) {
            case SINGLE:
                maxExplorerSteps = BehaviourConstants.MAX_EXPLORE_STEPS_SINGLE;
                exploredRate = BehaviourConstants.MAP_EXPLORED_RATE_SINGLE;
                heal = true;
                break;
            case DUEL:
                maxExplorerSteps = BehaviourConstants.MAX_EXPLORE_STEPS_DUEL;
                exploredRate = BehaviourConstants.MAP_EXPLORED_RATE_DUEL;
                heal = true;
                break;
            case JUNGLE:
                maxExplorerSteps = BehaviourConstants.MAX_EXPLORE_STEPS_JUNGLE;
                exploredRate = BehaviourConstants.MAP_EXPLORED_RATE_JUNGLE;
                heal = true;
                break;
        }
    }

    public synchronized void transformColonyToExplorers(QBeing me, BeingInterface bi) {
        if (!exploringFinished) {
            QBeing[] suppliers = colonyControl.getSuppliers(bi.getLocation(me));
            if (suppliers != null) {
                int i = (int) (Math.ceil(BehaviourConstants.KEEP_EATERS_WHEN_LACK_OF_FOOD * suppliers.length));
                if (i == 0)
                    i = 1;
                if (getBeingTypeCount(BeingType.TYPE_EXPLORER) + suppliers.length > BehaviourConstants.MAX_EXPLORER_COUNT) {
                    i = Math.max(i, suppliers.length - (BehaviourConstants.MAX_EXPLORER_COUNT - getBeingTypeCount(BeingType.TYPE_EXPLORER)));
                }
                while (i < suppliers.length) {
                    QBeing being = suppliers[i];
                    changeState(being.getState(), BeingState.S_EXPLORING, being, bi);
                    EventManager.handleBeingEvent(BeingEvent.E103, being, new StateMachineContext(being, bi));
                    i++;
                }
            }
        }
    }

    public synchronized float getExploredRate() {
        return exploredRate;
    }

    public synchronized Event colonyAction(BeingInterface bi, QBeing me) {
        return colonyControl.colonyAction(bi, me);
    }

    public synchronized void changeState(int oldState, int newState, QBeing me, BeingInterface bi) {
        decBeingTypeCount(me, oldState);
        incBeingTypeCount(me, newState);
        if (me.getState() == BeingState.S_PRODUCER) {
            colonySuplierRemoved(me.getLocation(), me);
            colonySuplierRemoved(bi.getLocation(me), me);
        }
        if (newState == BeingState.S_PRODUCER) {
            setColonySupplier(me, bi);
        }
        if (newState == BeingState.S_EXPLORING) {
            UniverseInfo.instance().setExplorerCenter(me, bi.getLocation(me));
        }
    }

    public synchronized Event moveTo(QBeing me, BeingInterface bi, Location location) {
        UniverseInfo.instance().updateBeingLocation(me, bi, bi.getLocation(me));
        UniverseInfo.instance().updateBeingLocation(me, bi, location);
        me.setLocation(location);
        return new Event(EventKind.ACTION_MOVE_TO, BeingUtil.stepToward(bi, me, location));
    }

    public synchronized Set<QBeing> getAllBeings() {
        Set<QBeing> set = new HashSet<QBeing>();
        for (Set<QBeing> gbs : beingsByType.values()) {
            set.addAll(gbs);
        }
        return set;
    }

    public synchronized void updateEnergy(QBeing me, float energy) {
        me.setEnergy(energy);
        colonyControl.updateupdateSupplierEnergy(me);
    }

    public synchronized int getMaxTurns() {
        return maxTurns;
    }

    public StrategyAutomata getAutomata() {
        return automata;
    }

    public synchronized void incBeingCount() {
        beingCount++;
    }

    public synchronized void setInitialState(QBeing me, BeingInterface bi) {
        checkExploringFinished(bi);
        float avail = UniverseInfo.instance().getAvailEnergy(me, new LocationInfo(bi.getPointInfo(me), me));
        if (avail > BehaviourConstants.MIN_FOOD_FOR_EATER * me.getMass()) {
            EventManager.handleBeingEvent(BeingEvent.E102, me, new StateMachineContext(me, bi));
            setColonySupplier(me, bi);
        } else if (!exploringFinished) {
            if (avail < BehaviourConstants.MIN_FOOD_FOR_EATER * me.getMass()) {
                EventManager.handleBeingEvent(BeingEvent.E103, me, new StateMachineContext(me, bi));
                transformColonyToExplorers(me, bi);
            } else if (getBeingTypeCount(BeingType.TYPE_EATER) < beingCount * BehaviourConstants.MIN_EATER_PERCENT ||
                    getBeingTypeCount(BeingType.TYPE_EXPLORER) >= BehaviourConstants.MAX_EXPLORER_COUNT) {
                EventManager.handleBeingEvent(BeingEvent.E102, me, new StateMachineContext(me, bi));
                setColonySupplier(me, bi);
            } else {
                EventManager.handleBeingEvent(BeingEvent.E102, me, new StateMachineContext(me, bi));
            }
        } else {
            Location loc = UniverseInfo.instance().findReachableEnergyPoint(me, bi, BehaviourConstants.CHANGE_FEED_PLACE_REQUIRED);
            if (loc != null) {
                me.setDestination(loc);
                EventManager.handleBeingEvent(BeingEvent.E104, me, new StateMachineContext(me, bi));
            } else {
                EventManager.handleBeingEvent(BeingEvent.E102, me, new StateMachineContext(me, bi));
                setColonySupplier(me, bi);
            }
        }
        changeState(0, me.getState(), me, bi);
    }


    public boolean isExploringFinished() {
        return exploringFinished;
    }

    public synchronized float getColonyForce(QBeing me) {
        Set<QBeing> suppliers = colonyControl.getSuppliersList(me.getLocation());
        float force = 0;
        if (suppliers != null) {
            for (QBeing being: suppliers) {
                force += being.getForce();
            }
        }
        force += UniverseInfo.instance().getSupport(me.getLocation());
        return force;
    }

    public synchronized void getSupport(final BeingInterface bi, final Location location, Float supportNeeded) {
        Set<QBeing> eaters = getBeingsByType(BeingType.TYPE_EATER);
        List<QBeing> potentialFighters = new ArrayList<QBeing>();
        for (QBeing being: eaters) {
            if (!being.getLocation().equals(location)) {
                if (being.getEnergy() >= being.getMass() * BehaviourConstants.SUPPORT_ENERGY) {
                    potentialFighters.add(being);
                }
            }
        }
        Collections.sort(potentialFighters, new Comparator<QBeing>() {
            public int compare(QBeing o1, QBeing o2) {
                float d1 = bi.distance(o1.getLocation(), location);
                float d2 = bi.distance(o2.getLocation(), location);
                return Float.compare(d1, d2);
            }
        });
        float force = 0;
        for (QBeing b: potentialFighters) {
            if (force >= supportNeeded) {
                break;
            }
            force += b.getForce();
            makeBeingWarrior(bi, b, location);
        }
    }

    private void makeBeingWarrior(BeingInterface bi, QBeing b, Location location) {
        UniverseInfo.instance().setSupportBeing(b, location);
        StateMachineContext ctx = new StateMachineContext(b, bi);
        b.setDestination(location);
        EventManager.handleBeingEvent(BeingEvent.E111, b, ctx);
    }
}
